/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2013-2023 Chaosdorf e.V.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package de.chaosdorf.meteroid

import android.os.Bundle
import androidx.activity.ComponentActivity
import androidx.activity.compose.setContent
import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.ArrowBack
import androidx.compose.material.icons.filled.Close
import androidx.compose.material.icons.twotone.Money
import androidx.compose.material3.BottomAppBar
import androidx.compose.material3.Button
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.ListItem
import androidx.compose.material3.Scaffold
import androidx.compose.material3.Text
import androidx.compose.material3.TextField
import androidx.compose.material3.TopAppBar
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.ui.Modifier
import androidx.compose.ui.tooling.preview.Preview
import androidx.hilt.navigation.compose.hiltViewModel
import androidx.lifecycle.ViewModel
import androidx.lifecycle.lifecycleScope
import androidx.lifecycle.viewModelScope
import androidx.lifecycle.viewmodel.compose.viewModel
import androidx.navigation.NavDestination.Companion.hierarchy
import androidx.navigation.compose.NavHost
import androidx.navigation.compose.composable
import androidx.navigation.compose.currentBackStackEntryAsState
import androidx.navigation.compose.navigation
import androidx.navigation.compose.rememberNavController
import coil.compose.AsyncImage
import dagger.hilt.android.AndroidEntryPoint
import dagger.hilt.android.lifecycle.HiltViewModel
import de.chaosdorf.mete.v1.MeteApiV1Factory
import de.chaosdorf.meteroid.icons.MeteroidIcons
import de.chaosdorf.meteroid.icons.twotone.WaterFull
import de.chaosdorf.meteroid.model.Drink
import de.chaosdorf.meteroid.model.DrinkRepository
import de.chaosdorf.meteroid.model.Server
import de.chaosdorf.meteroid.model.ServerId
import de.chaosdorf.meteroid.model.ServerRepository
import de.chaosdorf.meteroid.ui.theme.MeteroidTheme
import de.chaosdorf.meteroid.util.findBestIcon
import de.chaosdorf.meteroid.util.resolve
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.SharingStarted
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.collectLatest
import kotlinx.coroutines.flow.debounce
import kotlinx.coroutines.flow.flatMapLatest
import kotlinx.coroutines.flow.flowOf
import kotlinx.coroutines.flow.mapLatest
import kotlinx.coroutines.flow.stateIn
import kotlinx.coroutines.launch
import javax.inject.Inject
import kotlin.time.Duration.Companion.milliseconds

@AndroidEntryPoint
class MainActivity : ComponentActivity() {
  @Inject
  lateinit var accountPreferences: AccountPreferences

  @Inject
  lateinit var serverRepository: ServerRepository

  @Inject
  lateinit var drinkSyncHandler: DrinkSyncHandler

  override fun onCreate(savedInstanceState: Bundle?) {
    super.onCreate(savedInstanceState)

    val server = accountPreferences.state
      .mapLatest { it.server }
      .flatMapLatest {
        it?.let { serverRepository.getFlow(it) }
          ?: flowOf<Server?>(null)
      }

    lifecycleScope.launch {
      server.collectLatest {
        it?.let { server ->
          drinkSyncHandler.sync(server)
        }
      }
    }

    setContent {
      MeteroidTheme {
        AppRouter()
      }
    }
  }
}

@HiltViewModel
class AppViewModel @Inject constructor(
  private val accountPreferences: AccountPreferences,
  private val serverRepository: ServerRepository,
  private val drinkSyncHandler: DrinkSyncHandler
) : ViewModel() {
  val server: StateFlow<Server?> = accountPreferences.state
    .mapLatest { it.server }
    .flatMapLatest { serverId ->
      serverRepository.getAllFlow()
        .mapLatest { list ->
          list.firstOrNull { server -> server.serverId == serverId }
        }
    }.stateIn(viewModelScope, SharingStarted.WhileSubscribed(), null)

  init {
    viewModelScope.launch {
      server.collectLatest {
        it?.let { server ->
          drinkSyncHandler.sync(server)
        }
      }
    }
  }

  suspend fun selectServer(server: ServerId) {
    accountPreferences.setServer(server)
  }
}

@Composable
fun AppRouter(viewModel: AppViewModel = viewModel()) {
  val scope = rememberCoroutineScope()
  val server by viewModel.server.collectAsState()
  val navController = rememberNavController()

  LaunchedEffect(server) {
    if (server == null) {
      navController.navigate("accounts/list")
    }
  }

  Scaffold(topBar = {
    val navBackStackEntry by navController.currentBackStackEntryAsState()
    val currentDestination = navBackStackEntry?.destination
    if (currentDestination?.hierarchy?.any { it.route == "main" } == true) {
      TopAppBar(title = { Text("Meteroid") }, navigationIcon = {
        IconButton(onClick = { navController.navigate("accounts") }) {
          Icon(Icons.AutoMirrored.Default.ArrowBack, contentDescription = null)
        }
      })
    } else if (currentDestination?.hierarchy?.any { it.route == "accounts/list" } == true) {
      TopAppBar(title = { Text("Meteroid") }, navigationIcon = {
        IconButton(onClick = { navController.navigate("main") }) {
          Icon(Icons.Default.Close, contentDescription = null)
        }
      })
    }
  }, bottomBar = {
    val navBackStackEntry by navController.currentBackStackEntryAsState()
    val currentDestination = navBackStackEntry?.destination
    if (currentDestination?.hierarchy?.any { it.route == "main" } == true) {
      BottomAppBar(actions = {
        IconButton(onClick = { navController.navigate("main/purchase") }) {
          Icon(MeteroidIcons.TwoTone.WaterFull, contentDescription = null)
        }
        IconButton(onClick = { navController.navigate("main/deposit") }) {
          Icon(Icons.TwoTone.Money, contentDescription = null)
        }
      })
    }
  }) { padding ->
    NavHost(navController, startDestination = "main", Modifier.padding(padding)) {
      navigation(route = "accounts", startDestination = "accounts/list") {
        composable("accounts/list") { backStackEntry ->
          ServerListScreen(
            hiltViewModel(),
            onAddServer = { navController.navigate("accounts/addServer") },
            onSelectServer = {
              scope.launch {
                viewModel.selectServer(it)
                navController.navigate("main")
              }
            }
          )
        }
        composable("accounts/addServer") { backStackEntry ->
          AddServerScreen(
            hiltViewModel(),
            onAddServer = { navController.navigate("accounts/list") }
          )
        }
      }
      navigation(route = "main", startDestination = "main/purchase") {
        /*
        composable("users") { backStackEntry ->
          val viewModel = hiltViewModel<UserListViewModel>()
          UserListScreen(viewModel)
        }
        */
        composable("main/purchase") { backStackEntry ->
          DrinkListScreen(hiltViewModel())
        }
        composable("main/deposit") { backStackEntry ->

        }
      }
    }
  }
}

@HiltViewModel
class AddServerViewModel @Inject constructor(
  private val repository: ServerRepository
) : ViewModel() {
  val url = MutableStateFlow("")

  private suspend fun buildServer(
    id: ServerId,
    url: String
  ): Server? = try {
    val api = MeteApiV1Factory.newInstance(url)
    val manifest = api.getManifest()
    val icon = manifest?.findBestIcon()
    Server(
      id,
      manifest?.name,
      url,
      icon?.resolve(url)
    )
  } catch (_: Exception) {
    null
  }

  val server: StateFlow<Server?> = url
    .debounce(300.milliseconds)
    .mapLatest { buildServer(ServerId(-1), it) }
    .stateIn(viewModelScope, SharingStarted.WhileSubscribed(), null)

  suspend fun addServer() {
    val highestServerId = repository.getAll().maxOfOrNull { it.serverId.value } ?: 0
    val serverId = ServerId(highestServerId + 1)
    val server = buildServer(serverId, url.value)
    if (server != null) {
      repository.save(server)
    }
  }
}

@Composable
fun AddServerScreen(viewModel: AddServerViewModel, onAddServer: () -> Unit) {
  val scope = rememberCoroutineScope()
  val url by viewModel.url.collectAsState()
  val server by viewModel.server.collectAsState()

  Column {
    TextField(
      label = { Text("Server URL") },
      value = url,
      onValueChange = { viewModel.url.value = it }
    )

    Button(onClick = {
      scope.launch {
        viewModel.addServer()
        onAddServer()
      }
    }) {
      Text("Add Server")
    }

    server?.let { server ->
      Text(server.url)
      Text(server.name ?: "null1")
      AsyncImage(model = server.logoUrl, contentDescription = null)
    }
  }
}

@Preview
@Composable
fun ServerListScreen(
  viewModel: ServerListViewModel = viewModel(),
  onAddServer: () -> Unit = {},
  onSelectServer: (ServerId) -> Unit = {}
) {
  val servers by viewModel.servers.collectAsState()

  LazyColumn {
    items(servers) { server ->
      ListItem(
        headlineContent = { Text(server.name ?: server.url) },
        modifier = Modifier.clickable {
          onSelectServer(server.serverId)
        }
      )
    }
    item {
      ListItem(
        headlineContent = { Text("Add Server") },
        modifier = Modifier.clickable {
          onAddServer()
        }
      )
    }
  }
}

@Preview
@Composable
fun DrinkListScreen(
  viewModel: DrinkListViewModel = viewModel()
) {
  val drinks by viewModel.drinks.collectAsState()

  LazyColumn {
    items(drinks) { drink ->
      ListItem(headlineContent = { Text(drink.name) },
        supportingContent = { Text("${drink.volume}l · ${drink.price}€") })
    }
  }
}

@HiltViewModel
class DrinkListViewModel @Inject constructor(
  drinkRepository: DrinkRepository,
  accountPreferences: AccountPreferences
) : ViewModel() {
  private val serverId: Flow<ServerId?> = accountPreferences.state.mapLatest { it.server }
  val drinks: StateFlow<List<Drink>> = serverId.flatMapLatest {
    it?.let { serverId ->
      drinkRepository.getAllFlow(serverId)
    } ?: flowOf(emptyList())
  }.stateIn(viewModelScope, SharingStarted.WhileSubscribed(), emptyList())
}

@HiltViewModel
class ServerListViewModel @Inject constructor(
  serverRepository: ServerRepository
) : ViewModel() {
  val servers: StateFlow<List<Server>> = serverRepository.getAllFlow()
    .stateIn(viewModelScope, SharingStarted.WhileSubscribed(), emptyList())
}
